"""Implements a simple ros2 listener on `msg_type=String, topic="example/talker"`"""
import asyncio
from contextlib import suppress

import rclpy
from std_msgs.msg import String

from asyncio_for_robotics.ros2 import Sub, TopicInfo, auto_session

TOPIC = TopicInfo(msg_type=String, topic="example/talker")


async def main():
    # creates sub on the given topic
    sub = Sub(TOPIC.msg_type, TOPIC.topic, TOPIC.qos)
    # async for loop itterating every messages
    async for message in sub.listen_reliable():
        print(f"Received: {message.data}")


if __name__ == "__main__":
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(main()) # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        auto_session().close()
        rclpy.shutdown()



import asyncio
from contextlib import suppress

import rclpy
from example_interfaces.srv import AddTwoInts

import asyncio_for_robotics.ros2 as afor

TOPIC = afor.TopicInfo("add_two_ints", AddTwoInts)


async def fibo_client_loop():
    client = afor.Client(*TOPIC.as_arg())
    a: int = 1
    b: int = 0
    while 1:
        print(f"Requesting: {a} + {b}")
        try:
            response = await asyncio.wait_for(
                client.call(AddTwoInts.Request(a=a, b=b)), timeout=3
            )
        except TimeoutError:
            print("No response, requesting again")
            continue
        b = a
        a = response.sum
        print(f"Got: {a}")
        await asyncio.sleep(0.5)




if __name__ == "__main__":
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(fibo_client_loop())  # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        afor.auto_session().close()
        rclpy.shutdown()



import asyncio
import logging
import time
from contextlib import asynccontextmanager, suppress
from typing import Any, Awaitable, Callable, Coroutine, TypeVar

from asyncio_for_robotics.core.sub import BaseSub

try:
    from asyncio import timeout as john_timeout
except ImportError:
    from async_timeout import timeout as john_timeout

logger = logging.getLogger(__name__)

_T = TypeVar("_T")


async def soft_wait_for(coro: Awaitable[_T], timeout: float) -> _T | TimeoutError:
    """Awaits a coroutine with a timeout,
    if timeout occurs returns TimeoutError but does not raise it.

    Args:
        coro: coroutine to await
        timeout: timeout in seconds

    Returns:
        coroutine result or TimeoutError
    """

    timer = asyncio.create_task(asyncio.sleep(timeout))
    task = asyncio.ensure_future(coro)

    await asyncio.wait([timer, task], return_when=asyncio.FIRST_COMPLETED)

    if task.done():
        return task.result()
    else:
        return TimeoutError("afor.soft_wait_for")


@asynccontextmanager
async def soft_timeout(timeout: float):
    """
    Run an async block with a time limit, cancelling it on expiry.

    - Inside the block: awaited operations may be cancelled.
    - Outside the block: no exception propagates.

    Example:
        async with soft_timeout(1):
            await asyncio.sleep(2)  # interrupted after 1 second

        # Execution resumes here without raising CancelledError.

    Args:
        duration: Maximum time in seconds to allow the block to run.
    """
    timed_out = False
    try:
        async with john_timeout(timeout):
            yield lambda: timed_out
    except asyncio.CancelledError:
        timed_out = True
    except asyncio.TimeoutError:
        timed_out = True


class Rate(BaseSub[int]):
    def __init__(
        self, frequency: float, time_source: Callable[[], int] = time.time_ns
    ) -> None:
        """Provides a reliable, no drift rate, usable like an afor subscriber.

        Data returned is the time (in ns) at which the rate was suppose to fire.

        Posibilities:
            - wait_for_value returns the last tic time (ns)
            - wait_for_next and wait_for_new will wait for the next tic.
            - listen() executes every tic. However, if there is a queue, it
              skips and executes the latest.
            - listen_reliable() executes every tic and never misses any (queuing).

        Args:
            frequency: Frequency in Hz
            time_source: Source of time in ns
        """
        self.period: int= int(1e9 / frequency)
        super().__init__()
        self.time_source: Callable[[], int] = time_source
        self.periodic_task: asyncio.Task=self._initialize_task()

    def _initialize_task(self):
        async def periodic_coro():
            start_time = self.time_source()
            count = 0
            while 1:
                count += 1
                scheduled_time = start_time + count * (self.period)
                dt = scheduled_time - self.time_source()
                await asyncio.sleep(max(0, dt / 1e9))
                self._periodic_cbk(scheduled_time)

        return asyncio.create_task(periodic_coro())

    def _periodic_cbk(self, scheduled_time):
        self.input_data(scheduled_time)

    @property
    def name(self):
        return f"timer_{1e-9/self.period:.3f}Hz"

    def close(self):
        self.periodic_task.cancel()



import asyncio
from contextlib import suppress
import subprocess
from typing import IO, Any, Awaitable, List

from asyncio_for_robotics.core.sub import BaseSub


def make_afor_stdout_monitor(process: subprocess.Popen[str]) -> BaseSub[str]:
    """Creates an afor subscriber that returns every line of stdout of the given process."""
    loop = asyncio.get_running_loop()
    assert process.stdout is not None
    stdout: IO[str] = process.stdout
    afor_sub: BaseSub[str] = BaseSub()

    def reader():
        line = stdout.readline().strip()
        healthy = True
        if line:
            healthy = afor_sub.input_data(line)
        proc_ended = process.poll() is not None
        if proc_ended or not healthy:
            loop.remove_reader(stdout.fileno())

    loop.add_reader(stdout.fileno(), reader)
    return afor_sub


async def main():
    proc = subprocess.Popen(
        ["ping", "localhost"],
        stdout=subprocess.PIPE,
        text=True,
    )
    stdout_sub = make_afor_stdout_monitor(proc)
    async for line in stdout_sub.listen_reliable():
        print(f"I heard: \n   {line}")


if __name__ == "__main__":
    with suppress(KeyboardInterrupt):
        asyncio.run(main())
